﻿-- Uses: FlagRSPProtocol, FlagRSPConfigure, FlagRSPStatic, FlagRSP:printDebug

-- Public variables ------------------------------

TooltipHandler = {
    unit = "";
    unitName = "";

	guildMembers = {};
	friendEntries = {};
	ignoredList = {};

	updateFriendsTick = 0;
	updateGuildListTick = 0;
	updateIgnoredListTick = 0;
};



-- Constant variables ----------------------------

-- Colors in the form of Alpha Red Green Blue in Hex (8 symbols, AARRGGBB)
local COLOR_RPTITLE       = "FFFFD100";
local COLOR_CURRENTDESC   = "FFFFD100";
local COLOR_FLAGRP        = "FFFFFF80";
local COLOR_FLAGCHAR      = "FFFFFF80";
local COLOR_INGAMENAME    = "FF909090";
--local COLOR_RPTITLEFL = "FFFFFF7F";

-- Localized strings for the tooltip use now Blizzard's texts from GlobalStrings.lua
local L_LEVEL = LEVEL; -- "Level" in the tooltip (MUST BE EXACT!)

local L_MINIONLINES = {}; -- Pet detection for pet owner search in the tooltip (MUST BE EXACT!)
L_MINIONLINES[0] = gsub(UNITNAME_SUMMON_TITLE1, "%%s", "(.+)"); -- Pet
L_MINIONLINES[1] = gsub(UNITNAME_SUMMON_TITLE3, "%%s", "(.+)"); -- Minion
L_MINIONLINES[2] = gsub(UNITNAME_SUMMON_TITLE5, "%%s", "(.+)"); -- Companion

local L_MINION = {}; -- "%'s whatever"
L_MINION[0] = UNITNAME_SUMMON_TITLE1; -- Pet;
L_MINION[1] = UNITNAME_SUMMON_TITLE3; -- Minion
L_MINION[2] = UNITNAME_SUMMON_TITLE5; -- Companion

local L_AKA = "aka"; -- "also known as"
local L_CURRENT = "Currently";

local MAXIMAL_TT_WIDTH = 250;

local orig_GameTooltip_OnTooltipSetUnit = GameTooltip:GetScript("OnTooltipSetUnit");


-- Event handling --------------------------------

function TooltipHandler:OnLoad( frame )
	GameTooltip:SetScript("OnTooltipSetUnit", function(widget)
        orig_GameTooltip_OnTooltipSetUnit( widget ); -- original OnTooltipSetUnit of the GameTooltip
        TooltipHandler:OnTooltipSetUnit( widget );
        end);
end

function TooltipHandler:OnEvent_Init( arg )
    -- TooltipHandler:BuildSkinnableTable();
end

function TooltipHandler:OnUpdate( elapsed ) -- XXX belongs somewhere else?
	local curTime = GetTime();

	if ( curTime > self.updateFriendsTick ) then
		self:UpdateFriendEntries();
		self.updateFriendsTick = curTime + 30;
	end

	if ( curTime > self.updateGuildListTick ) then
		self:UpdateGuildMembers();
		self.updateGuildListTick = curTime + 300;
	end

	if ( curTime > self.updateIgnoredListTick ) then
		self:UpdateIgnoredList();
		self.updateIgnoredListTick = curTime + 30;
	end
end

function TooltipHandler:OnTooltipSetUnit( widget )    
    self.unit = "";
    self.unitName = "";
    local tName, tUnit = GameTooltip:GetUnit();
    -- print("TooltipHandler:OnTooltipSetUnit(  ) fired -> " .. tUnit);
    if tUnit ~= nil then
        if strfind(tUnit,"^party") or strfind(tUnit, "^raid") or strfind(tUnit, "^focus")
                -- or strfind(tUnit, "^player") or strfind(tUnit, "^pet")
                or tUnit == "mouseover"
                then
            -- print("TooltipHandler:OnTooltipSetUnit(  ) working -> ok");
            local request = FlagRSPProtocol:Request(tUnit, "ott");
            FlagRSPProtocol:Update(tUnit, tUnit, request);
        end
    end
    self:UpdateTooltip( );
end



-- Manipulate tooltip ----------------------------

function TooltipHandler:UpdateTooltip( reset )
    if FlagRSP.selfcast then
        return;
    end
    local tooltipBase = "GameTooltip";	
	local tooltipUnitName, tooltipUnit = getglobal(tooltipBase):GetUnit();
    tooltipUnit = tooltipUnit or "mouseover";

    -- Valid individual: only creatures, characters, npc, ...
	if UnitName(tooltipUnit) == nil then
		return;
	end

    -- No modified tooltip
    if FlagRSPConfigure:GetTooltipStyle() == "none" then
        return;
    end
    if reset then
        self.unit = "";
        self.unitName = "";
    end
    if self.unit == tooltipUnit or self.unitName == tooltipUnitName then
        return;
    else
        self.unit = tooltipUnit;
        self.unitName = tooltipUnitName;
    end

    -- Valid individual type (character, npc, pet, ...)
    local targetType = "";
    if ( UnitIsPlayer(tooltipUnit) ) then
        targetType = "player";
    elseif ( not UnitIsPlayer(tooltipUnit)) and (UnitPlayerControlled(tooltipUnit) ) then
		targetType = "pet";
    else
        targetType = "npc";
    end

    -- Modify tooltip dependend on tooltip style
    -- FlagRSP tooltip
    if ( FlagRSPConfigure:GetTooltipStyle() == "normal" ) then
        self:UpdateTooltipRP( tooltipUnit, targetType );
        self:UpdateTooltipLevel( tooltipUnit, tooltipBase )

    -- Slightly modified tooltip (only appending)
    elseif ( FlagRSPConfigure:GetTooltipStyle() == "light" ) then
        self:UpdateTooltipAppend( tooltipUnit, targetType );

    end

    -- Fix line width
    for i = 1, getglobal(tooltipBase):NumLines() do
        getglobal(tooltipBase .. "TextLeft" .. i):SetWidth(MAXIMAL_TT_WIDTH);
    end
end

function TooltipHandler:UpdateTooltipLevel( tooltipUnit, tooltipBase )
    local tt = getglobal(tooltipBase);

    -- Replace alternatve level
    if ( FlagRSPConfigure:GetLevelDisplay() ~= "none" ) then
        local level = L_LEVEL .. " " .. UnitLevel(tooltipUnit);
        for i = 1, tt:NumLines() do
            local textLine = getglobal(tooltipBase .. "TextLeft" .. i);
            local text = textLine:GetText();
            if ( strfind(text, L_LEVEL) ) then
                local lvlchange = "";
                if ( FlagRSPConfigure:GetLevelDisplay() == "color" ) then
                    lvlchange = TooltipHandler:GetAlternateLevelText(tooltipUnit);
                elseif ( FlagRSPConfigure:GetLevelDisplay() == "relative") then
                    lvlchange = TooltipHandler:GetAlternativeEnemyLevelText(tooltipUnit);
                end
                if lvlchange ~= nil and lvlchange ~= "" then
                    text = gsub(text, level, lvlchange);
                end

                textLine:SetText(text);
                break;
            end
        end
    end
    tt:Show();
end

function TooltipHandler:UpdateTooltipRP( tooltipUnit, targetType )
    local tooltipBase = "GameTooltip";
	if tooltipUnit == "" or tooltipUnit == nil then
		tooltipUnit = "mouseover";
	end

    -- Get state to the unit
    local tt = getglobal("GameTooltip");
    local modifier = "";
    if (UnitIsPlayer(tooltipUnit)) then
        if (self:IsPlayerIgnored(UnitName(tooltipUnit))) then
			modifier = "ignored";
        elseif (not FlagRSPConfigure:GetNameDisplay() or TooltipHandler:PlayerIsKnown(UnitName(tooltipUnit))) then
			modifier = "known";
		else
			modifier = "unknown";
		end
    elseif UnitPlayerControlled(tooltipUnit) then
        if (not FlagRSPConfigure:GetNameDisplay() or TooltipHandler:PlayerIsKnown(UnitName(tooltipUnit))) then
			modifier = "known";
		else
			modifier = "unknown";
		end
	end

    -- Color the tooltip based on modifier
    local r, g, b = TooltipHandler:GetTooltipBGColor(tooltipUnit);
    tt:SetBackdropColor(r, g, b);	
    if (modifier == "ignored") then
        tt:SetBackdropColor(0.5,0,0);
    end

    -- Save original tooltip contents
    local tooltipLines = {};
    for i = 1, tt:NumLines() do
        local text = getglobal(tooltipBase .. "TextLeft" .. i):GetText();
        local r, g, b = getglobal(tooltipBase .. "TextLeft" .. i):GetTextColor();
        text = text or "";
        text = "|cFF" .. FlagRSPStatic:GetHex(r, g, b) .. text .. "|r";
        tinsert(tooltipLines,text);
    end

    -- Tear down and rebuild tooltip
    if targetType == "player" then
        local unitName, unitRealm = UnitName(tooltipUnit);

        -- Collect relevant info ("raw")
        local ingameName     = unitName; -- Game Name
        local title          = self:GetGameTitle( tooltipUnit ); -- Game title
        local guild, rank, _ = self:GetGameGuild( tooltipUnit ); -- Game guild , rank, rankIndex
        local level          = self:GetLevel( tooltipUnit ); -- Game level
        local race           = self:GetRace( tooltipUnit ); -- Game race
        local class          = self:GetClass( tooltipUnit ); -- Game class
        local faction        = self:GetFaction( tooltipUnit ); -- Game faction
        local fullname       = self:GetFullname( unitName, unitRealm ); -- RP Name
        local rptitle        = self:GetRPTitle( unitName, unitRealm ); -- RP Title
        local rpflag         = self:GetRPFlag( unitName, unitRealm ); -- RP RP Flag
        local charflag       = self:GetCharFlag( unitName, unitRealm ); -- RP Char Flag
        local friendstate    = self:GetFriendState( unitName, unitRealm ); -- Friendstate
        -- local currentDesc    = self:GetCurrently( unitName, unitRealm );
        local flagaddon      = self:GetFlagAddon( unitName, unitRealm );

        -- Check configurations, just set "" for hiding
        if ( unitRealm ~= nil ) then -- Foreign realm
            ingameName = GetUnitName(tooltipUnit, true);
            if ( fullname == "" ) then
                fullname = GetUnitName(tooltipUnit);
            end
        end
        if ( FlagRSPConfigure:GetNameDisplay()) then -- Hide name
            if ( not self:PlayerIsKnown(unitName) ) then
                fullname = FlagRSP_Locale_Unknown;
            end
        end
        if ( fullname == "" ) then -- Use ingameName if fullname empty
                fullname = ingameName;
                ingameName = "";
        end
        if ( strfind(ingameName, fullname) ) then -- Fullname is contained in ingame name
            ingameName = "";
        end
        if ( not FlagRSPConfigure:GetShowUnitName() ) then -- Hide ingame name
            ingameName = "";
        end
        if ( not FlagRSPConfigure:GetPVPRanks() ) then -- Hide game title
            title = "";
        end

        -- Prepare lines for the tooltip (additional texts, colors, ...)
        local ingameNameLine       = self:DecorateIngameName( ingameName );
        local titleLine            = self:DecorateGameTitle( title );
        local guildLine            = self:DecorateGameGuild( guild, rank, nil );
        local lvlLine              = self:DecorateCharInfo( level, race, class );
        local factionLine          = self:DecorateFaction( faction );
        local fullnameLine         = self:DecorateFullname( tooltipUnit, fullname );
        local rptitleLine          = self:DecorateRPTitle( rptitle );
        local rpflagLine           = self:DecorateRPFlag( rpflag );
        local charflagLine         = self:DecorateCharFlag( charflag );
        local friendstateLine      = self:DecorateFriendstate( unitName, friendstate, unitRealm );
        local friendstateGuildLine = self:DecorateFriendstateGuild( guild, unitRealm );
        -- local currentLine          = self:DecorateCurrent( currentDesc );
        local addonLine            = self:DecorateAddon( flagaddon );

        local lines = {};
        self:AddTooltipTable( lines, fullname    , fullnameLine );
        self:AddTooltipTable( lines, ingameName  , ingameNameLine );
        self:AddTooltipTable( lines, rptitle     , rptitleLine );
        self:AddTooltipTable( lines, guild       , guildLine );
        self:AddTooltipTable( lines, "lvlLine"   , lvlLine );
        self:AddTooltipTable( lines, faction     , factionLine );
        self:AddTooltipTable( lines, title       , titleLine );
        self:AddTooltipTable( lines, rpflag      , rpflagLine );
        self:AddTooltipTable( lines, charflag    , charflagLine );
        self:AddTooltipTable( lines, friendstate , friendstateLine );
        self:AddTooltipTable( lines, guild       , friendstateGuildLine );
        -- self:AddTooltipTable( lines, currentDesc , currentLine );
        -- self:AddTooltipTable( lines, flagaddon   , addonLine );

        self:RemoveReplacedLine( tooltipLines, "ingamename", unitName );
        self:RemoveReplacedLine( tooltipLines, "guild"     , guild );
        self:RemoveReplacedLine( tooltipLines, "level"     , level );
        self:RemoveReplacedLine( tooltipLines, "faction"   , faction );

        -- Clear and rebuild tooltip
        local original_lines = tt:NumLines();

        for i = 1, tt:NumLines() do
            getglobal(tooltipBase .. "TextLeft" .. i):SetTextColor(1,1,1,1);
            getglobal(tooltipBase .. "TextLeft" .. i):SetText(nil);
            -- getglobal(tooltipBase .. "TextLeft" .. i):SetWidth(MAXIMAL_TT_WIDTH);
        end

        self:RebuildTooltip( tooltipBase, original_lines, 1, lines );
        self:RebuildTooltip( tooltipBase, original_lines, 0, tooltipLines );

    elseif targetType == "pet" then
        -- Change first line color
        local r, g, b = GameTooltip_UnitColor(tooltipUnit);
        getglobal(tooltipBase .. "TextLeft1"):SetTextColor(r, g, b, 1.0);

        local owner = self:GetPetOwner(tooltipLines);
        local ownerknown = self:PlayerIsKnown(owner);

        owner = FlagRSPStatic:RemoveColourCode(owner);

        -- Player knows himself and his own pets. Always.
        if owner == select(1,UnitName("player")) then
            modifier = "known";
            ownerknown = true;
        end

        local unitName, _ = UnitName(tooltipUnit);
        for i = 1, tt:NumLines() do
            local ttLeft = getglobal( tooltipBase .. "TextLeft" .. i);
            local text = ttLeft:GetText();
            text = FlagRSPStatic:RemoveColourCode(text);
            if ( FlagRSPConfigure:GetNameDisplay()) then
                -- Hide unknown pet name
                if ( text == unitName and modifier == "unknown") then -- pet name
                    text = ttLeft:GetText();
                    if ( strfind(unitName, "-") ) then -- Fix for warlock pet names with dash
                        text = gsub(text,"-","_");
                        unitName = gsub(unitName,"-","_");
                        text = gsub(text,"_","-");
                    end
                    text = gsub(text, unitName, FlagRSP_Locale_UnknownPet);
                    ttLeft:SetText(text);

                -- Hide unknown owner
                elseif ( strfind(text, owner) and not ownerknown) then -- owner line
                    text = ttLeft:GetText();
                    text = gsub(text, owner, FlagRSP_Locale_Unknown);
                    ttLeft:SetText(text);
                end
            end
        end

    elseif targetType == "npc" then
        -- Change first line color
        local r, g, b = GameTooltip_UnitColor(tooltipUnit);
        getglobal(tooltipBase .. "TextLeft1"):SetTextColor(r, g, b, 1.0);

        -- Add RP Stuff
        local unitName, unitRealm = UnitName(tooltipUnit);

        local friendstate = self:GetFriendState(unitName, unitRealm);

        friendstate = self:DecorateFriendstate(unitName, friendstate, nil)

        self:AddTooltipLine(friendstate);
    end

	tt:Show();

end

function TooltipHandler:UpdateTooltipAppend( tooltipUnit, targetType )
    local unitName, unitRealm = UnitName(tooltipUnit);
    
    local unitKey = unitName;
    if unitRealm ~= nil and unitRealm ~= FlagRSP.rName then
        unitKey = unitName .. "-" .. unitRealm;
    end

    if targetType == "player" then

        local fullname    = self:GetFullname(unitName, unitRealm);
        local rptitle     = self:GetRPTitle(unitName, unitRealm);
        local rpflag      = self:GetRPFlag( unitName, unitRealm );
        local cstatus     = self:GetCharFlag( unitName, unitRealm );
        local friendstate = self:GetFriendState(unitName, unitRealm);
        local guild, _, _ = self:GetGameGuild( tooltipUnit ); -- Game guild , rank, rankIndex

        fullname    = self:DecorateFullnameAka( tooltipUnit, fullname );
        rptitle     = self:DecorateRPTitle( rptitle );
        rpflag      = self:DecorateRPFlag( rpflag );
        cstatus     = self:DecorateCharFlag( cstatus )
        friendstate = self:DecorateFriendstate( unitName, friendstate, unitRealm )
        guild       = self:DecorateFriendstateGuild( guild, unitRealm );

        if ( fullname ~= "" or rptitle ~= "" or rpflag ~= "" or cstatus ~= "" or friendstate ~= "" ) then
            self:AddTooltipLine("  ");
        end

        self:AddTooltipLine(fullname);
        self:AddTooltipLine(rptitle);
        self:AddTooltipLine(rpflag);
        self:AddTooltipLine(cstatus);
        self:AddTooltipLine(friendstate);
        self:AddTooltipLine(guild);

    elseif targetType == "npc" then
        local friendstate = self:GetFriendState(unitName, unitRealm);

        friendstate = self:DecorateFriendstate( unitName, friendstate, nil )

        self:AddTooltipLine(friendstate);

    end

    GameTooltip:Show();

end

function TooltipHandler:IsContained( tooltipBase, textline )
    if tooltipBase == nil or tooltipBase == "" then
        tooltipBase = "GameTooltip";
    end

    textline = textline or "";
    textline = FlagRSPStatic:RemoveColourCode(textline);

    local temp;
    local isContained = false;
    for i = 1, getglobal(tooltipBase):NumLines() do
		temp = getglobal(tooltipBase .. "TextLeft"..i):GetText();
        temp = FlagRSPStatic:RemoveColourCode(temp);
		if (textline == temp) then
			isContained = true
            break;
		end
	end

    return isContained;
end

function TooltipHandler:AddTooltipLine( textline )
    local tooltipBase = "GameTooltip";

	local temp;
    local tt = getglobal(tooltipBase);
	local alreadyIn = self:IsContained( tooltipBase, textline );

	if (not alreadyIn) then
		tt:AddLine(textline, 1, 1, 1);
	end

end

function TooltipHandler:AddTooltipTable( lines, var, text )
    if ( var ~= nil and var ~= "" ) then
        tinsert( lines, text );
    end
end

function TooltipHandler:RemoveReplacedLine( lines, caption, value )

    if ( value == "" ) then
        return;
    end

    value = FlagRSPStatic:RemoveColourCode(value);

    for i = 1, table.getn(lines) do

        if caption == "ingamename" then
            if ( strfind(lines[i], value) ) then
                lines[i] = "";
            end
        elseif caption == "guild" then
            local line = FlagRSPStatic:RemoveColourCode(lines[i]);
            value = FlagRSPStatic:RemoveColourCode(value);
            if ( strfind(line, value) or strfind(value, line) ) then
                lines[i] = "";
            end
        elseif caption == "level" then
            if ( strfind(lines[i], L_LEVEL) ) then
                lines[i] = "";
            end
        elseif caption == "faction" then
            local line = FlagRSPStatic:RemoveColourCode(lines[i]);
            if ( line == value ) then
                lines[i] = "";
            end
        end
    end

end

function TooltipHandler:RebuildTooltip( tooltipBase, original_lines, start, lineTable )

    if ( start == nil or tonumber(start) == nil or tonumber(start) < 0) then
        start = 1;
    end

    local tt = getglobal(tooltipBase);

    if ( start == 0 ) then
        start = tt:NumLines();
    end

    local current_line = start;
    for t = 1, table.getn(lineTable) do
        local ttLeft = getglobal(tooltipBase .. "TextLeft" .. current_line);

        if ( lineTable[t] ~= nil and lineTable[t] ~= "" ) then
            if not self:IsContained( tooltipBase, lineTable[t] ) then
                if ( current_line <= original_lines ) then
                    ttLeft:SetText(lineTable[t]);                    
                else
                    self:AddTooltipLine( lineTable[t] );
                end
                current_line = current_line + 1;
            end

        end
    end

end



-- Get Game and RP Data --------------------------

function TooltipHandler:GetGameTitle( playerID )
	local Rank = "";

	-- capture self-made playerID for Friendlist
	if string.sub(playerID,1,5) == "index" then
		local iname = self:GetCharacterName(playerID, "unit");
		if ( self:GetFriendInfo("rank",iname) ~= nil ) then
			Rank = self:GetFriendInfo("rank",iname);
		end
	else
		local name = UnitName(playerID);
		if ( UnitPVPName(playerID) ~= name and name ~= nil and UnitPVPName(playerID) ~= nil ) then
			Rank = UnitPVPName(playerID) or "";
            Rank = gsub(Rank, name, "");
            Rank = gsub(Rank, ",", "");
            Rank = strtrim(Rank);
		end
	end

	return Rank;
end

function TooltipHandler:GetGameGuild( playerID )

	local guild = "";
    local rank = "";
    local rankIndex = -1;

	-- capture self-made playerID for Friendlist
	if string.sub(playerID,1,5) == "index" then
		local iname = self:GetCharacterName(playerID, "unit");
		guild = TooltipHandler:GetFriendInfo("guild", iname);
	else
		guild, rank, rankIndex = GetGuildInfo(playerID);
	end
	
    guild = guild or "";
    rank = rank or "";
    rankIndex = rankIndex or -1;

	return guild, rank, rankIndex;
end

function TooltipHandler:GetLevel( playerID )
	local Level = "";

	if string.sub(playerID,1,5) == "index" then
		local iname = self:GetCharacterName(playerID, "unit");
		Level = TooltipHandler:GetFriendInfo("level", iname);
	else
		if UnitLevel(playerID) >= 1 then
			Level = UnitLevel(playerID);
		else
			Level = "??";
		end
	end

	return Level;
end

function TooltipHandler:GetRace( playerID )
	local Race = "";

	-- capture self-made playerID for Friendlist
	if string.sub(playerID,1,5) == "index" then
		local iname = self:GetCharacterName(playerID, "unit");
		Race = TooltipHandler:GetFriendInfo("race", iname);
	else
		if UnitRace(playerID) ~= nil then
			Race = select(1, UnitRace(playerID));
		end
	end
	
	if ( Race == nil ) then
		Race = "";
	end

	return Race;
end

function TooltipHandler:GetClass( playerID )
	local Class = "";

	-- capture self-made playerID for Friendlist
	if string.sub(playerID,1,5) == "index" then
		local iname = self:GetCharacterName(playerID, "unit");
		Class = TooltipHandler:GetFriendInfo("class", iname);
	else
		if UnitClass(playerID) ~= nil then
			-- Lets the class be colored on the tooltip.
			if ( FlagRSPConfigure:GetClassColor() ) then
				local class, file = UnitClass(playerID);
				local color;
				color = RAID_CLASS_COLORS[file];
				
				Class = "|cFF" .. FlagRSPStatic:GetHex(color.r, color.g, color.b);
				Class = Class .. class .. "|r";
			else
				Class = UnitClass(playerID);
		end
	end
		if UnitCreatureType(playerID) ~= nil and UnitCreatureType(playerID) ~= "" and not UnitIsPlayer(playerID) and not UnitPlayerControlled(playerID) then
			Class = UnitCreatureType(playerID);
		end
		if UnitCreatureFamily(playerID) ~= nil and UnitCreatureFamily(playerID) ~= "" then
			Class = UnitCreatureFamily(playerID);
		end
	end

	return Class;
end

function TooltipHandler:GetFaction( playerID )
	local faction = "";

	-- capture self-made playerID for Friendlist
	if string.sub(playerID,1,5) == "index" then
		-- local iname = self:GetCharacterName(playerID, "unit");
		-- Race = TooltipHandler:GetFriendInfo("race", iname);
	else
		if UnitFactionGroup(playerID) ~= nil then
			faction = select(2, UnitFactionGroup(playerID));
		end
	end
	
	if ( faction == nil ) then
		faction = "";
	end

	return faction;
end

function TooltipHandler:GetFullname( name, realm )
	local fullname = "";

	if TooltipHandler:GetFlagInfo("Fullname", name, realm) ~= "" then
		fullname = TooltipHandler:GetFlagInfo("Fullname", name, realm);
	elseif TooltipHandler:GetFriendInfo("fullname", name, realm) ~= "" and TooltipHandler:GetFriendInfo("fullname", name, realm) ~= nil then
		fullname = TooltipHandler:GetFriendInfo("fullname", name, realm);
	end

	return fullname;
end

function TooltipHandler:GetRPTitle( name, realm )
	local title = "";

	if TooltipHandler:GetFlagInfo("Title", name, realm) ~= "" then
		title = TooltipHandler:GetFlagInfo("Title", name, realm);
	elseif TooltipHandler:GetFriendInfo("title", name, realm) ~= nil then
		title = TooltipHandler:GetFriendInfo("title", name, realm);
	end

	return title;
end

function TooltipHandler:GetRPFlag( name, realm )
	local rpflag = TooltipHandler:GetFlagInfo("RPFlag", name, realm);
	
	if tonumber(rpflag) and rpflag >= 1 then
		return rpflag;
	else
		return 0;
	end
end

function TooltipHandler:GetCharFlag( name, realm )
    local cstatus = TooltipHandler:GetFlagInfo("CharFlag", name, realm);
	if tonumber(cstatus) and cstatus >= 1 then
		return cstatus;
	else
		return 0;
	end
end

function TooltipHandler:GetFriendState( playerID, realm )
    return self:GetFriendInfo("friendstate", playerID, realm);
end

function TooltipHandler:GetPetOwner( tooltipLines )
	local ownerName = "";

	for t=1, 8 do
		if tooltipLines[t] ~= nil then
			local i = 0;
			while L_MINIONLINES[i] ~= nil do

				local s, e, match = string.find(tooltipLines[t], L_MINIONLINES[i]);

				if match ~= nil and match ~= "" then
					ownerName = match;
                    break;
				end
				i = i + 1;
			end
		end
	end

	return ownerName;
end

function TooltipHandler:GetCurrently( unitName, unitRealm )
    local MAXLENGTH = 42;
    local current = "";

    current = TooltipHandler:GetFlagInfo("Current", unitName, unitRealm);
    if ( current == nil or current == "" ) then
        current = TooltipHandler:GetFlagInfo("Desc", unitName, unitRealm);
    end
    if current and current ~= "" then
        if ( strlen(current) > MAXLENGTH ) then
            local pos = strfind(current, "\n");
            if pos == nil or pos > MAXLENGTH then
                current = strsub(current, 1, MAXLENGTH) .. "...";
            elseif pos then
                current = strsub(current, 1, pos) .. " ... ";
            end
        end
        current = gsub(current, "\n", "");
    end

    return current;
end

function TooltipHandler:GetFlagAddon( unitName, unitRealm )
    local flagaddon, flagversion;
    local flagaddons;
    flagaddons = FlagHandler:GetFlag( "FlagAddon", unitName, unitRealm );
    
    return flagaddons, flagversion;
end


-- Decorate data for lines -----------------------

function TooltipHandler:DecorateIngameName( ingameName )
	return "|c"..COLOR_INGAMENAME.."(" .. ingameName .. ")|r";
end

function TooltipHandler:DecorateGameTitle( title )
	return UNIT_NAME_PLAYER_TITLE .. ": " .. title;
end

function TooltipHandler:DecorateGameGuild( guild, rank, rankIndex )
    guild = strtrim(guild or "");
    rank  = strtrim(rank or "");

    -- Build line
	local guildText = "";

	if FlagRSPConfigure:GetGuildRank() and guild ~= "" and rank ~= ""then
        local text = "%GUILDNAME%, %GUILDRANK%";
        -- text = "%GUILDRANK% " .. FlagRSP_Locale_Of .. " %GUILDNAME%"; -- Old style, uncomment for use
        text = gsub(text, "%%GUILDNAME%%", guild)
        text = gsub(text, "%%GUILDRANK%%", rank)
        guildText = text;
	elseif guild ~= "" then
        local text = "%GUILDNAME%";
        text = gsub(text, "%%GUILDNAME%%", guild)
        guildText = text;
	end

    -- Check Hiding
    if ( FlagRSPConfigure:GetGuildDisplay() == "known" ) then
		if not Friendlist:IsFriend(guild) then
			guildText = "";
		end
    elseif ( FlagRSPConfigure:GetGuildDisplay() == "none" ) then
        guildText = "";
    end

	return guildText;
end

function TooltipHandler:DecorateCharInfo( level, race, class )
    level = level or "?";
    race = race or "?";
    class = class or "?";

    -- Example: Level 90 - Nightelf, Hunter (Player)
    local text = L_LEVEL .." %LEVEL% - %RACE%, %CLASS% (" .. STATUS_TEXT_PLAYER .. ")";
    -- text = L_LEVEL .." %LEVEL% %RACE%  %CLASS% (" .. STATUS_TEXT_PLAYER .. ")"; -- Old style, uncomment for use
    text = gsub(text, "%%LEVEL%%", level)
    text = gsub(text, "%%RACE%%", race)
    text = gsub(text, "%%CLASS%%", class)

    return text;
end

function TooltipHandler:DecorateFaction( faction )
	return faction;
end

function TooltipHandler:DecorateFullname( tooltipUnit, fullname )
    tooltipUnit = tooltipUnit or "mouseover";
    fullname = fullname or "";

    local text = "";

    if fullname ~= "" then
        local r, g, b = self:GetTooltipNameColor(tooltipUnit);
        local colorName = FlagRSPStatic:GetHex(r, g, b);
        text = "|cFF" .. colorName .. fullname .. "|r"
    end

    return text;
end

function TooltipHandler:DecorateFullnameAka( tooltipUnit, fullname )
    fullname = fullname or "";

    local text = "";

    if fullname ~= "" then
        text = self:DecorateFullname( tooltipUnit, fullname );
        text = L_AKA .. " " .. text;
    end

    return text;
end

function TooltipHandler:DecorateRPTitle( rptitle )
    rptitle = rptitle or "";

    local text = "";
    if rptitle ~= "" then
        text = "|c" .. COLOR_RPTITLE .. rptitle .. "|r"
    end

    return text;
end

function TooltipHandler:DecorateRPFlag( rpflag )
	local s = "";
	
	if tonumber(rpflag) ~= nil and rpflag >= 1 then
		s = s .. FlagRSPStatic:GetRPFlagText( rpflag, true );
	end
	
	return s;
end

function TooltipHandler:DecorateCharFlag( cstatus )
	cstatus = cstatus or 0;
	
	local charstatus = "";

	charstatus = FlagRSPStatic:GetCharStateText( cstatus, true );
	if charstatus ~= "" then
		charstatus = FlagRSP_Locale_CharStatusLine .. charstatus;
	end
	
	return charstatus;
end

function TooltipHandler:DecorateFriendstate( unitName, friendstate, realm )
    unitName = unitName or "";
    friendstate = friendstate or "";

    local text = "";

    if ( unitName ~= "" and friendstate ~= "" ) then
        text = "|cFF" .. TooltipHandler:GetFriendInfo("friendcolor", unitName, realm) .. friendstate .. "|r"
    end
    return text;
end

function TooltipHandler:DecorateFriendstateGuild( guildName , unitRealm )
    guildName = guildName or "";
    unitRealm = unitRealm or "";
    
    local text = "";
    
    if Friendlist:IsFriend(guildName) then
        local r,g,b,t = Friendlist:GetFriendstateInfo(Friendlist:GetFriendstate(guildName));
        -- print(r .. "." .. g .."."..b.." : " ..t);
        text = "|cFF"..FlagRSPStatic:GetHex( r, g, b )..Friendlist:GetFriendstateText(guildName).."|r";
    end
    
    return text;
end

function TooltipHandler:DecorateCurrent( currentDesc )
	return "|c"..COLOR_CURRENTDESC .. L_CURRENT .. ":|r \n" .. currentDesc;
end

--[[

- TooltipHandler:GetAlternativeEnemyLevelText(playerID)

- Returns alternative level information for enemies.

]]--
function TooltipHandler:GetAlternativeEnemyLevelText( playerID )
	local Type, prefix;
	local level;

	if string.sub(playerID,1,5) == "index" then
		local iname = self:GetCharacterName(playerID, "unit");
		level = TooltipHandler:GetFriendInfo("level", iname);
	else
		level = UnitLevel(playerID);
	end

	if level ~= nil and level ~= "" and level ~= 0 then
		Diff = level - UnitLevel("player");

		if string.sub(playerID,1,5) ~= "index" and UnitClassification(playerID) == "elite" then
			Diff = Diff + 5;
		end

		local r, g, b;
		if ( level == -1 or Diff >= 7 ) then
			r = QuestDifficultyColors["impossible"].r;
			g = QuestDifficultyColors["impossible"].g;
			b = QuestDifficultyColors["impossible"].b;
		elseif ( Diff >= 4 ) then
			r = QuestDifficultyColors["verydifficult"].r;
			g = QuestDifficultyColors["verydifficult"].g;
			b = QuestDifficultyColors["verydifficult"].b;
		elseif ( Diff >= -1 ) then
			r = QuestDifficultyColors["difficult"].r;
			g = QuestDifficultyColors["difficult"].g;
			b = QuestDifficultyColors["difficult"].b;
		elseif ( -Diff <= GetQuestGreenRange() ) then
			r = QuestDifficultyColors["standard"].r;
			g = QuestDifficultyColors["standard"].g;
			b = QuestDifficultyColors["standard"].b;
		else
			r = QuestDifficultyColors["trivial"].r;
			g = QuestDifficultyColors["trivial"].g;
			b = QuestDifficultyColors["trivial"].b;
		end

		prefix = "|cFF" .. FlagRSPStatic:GetHex(r, g, b);

		if ( level == -1 or Diff >= 10 ) then
			Type = prefix .. FlagRSP_Locale_Impossible;
		elseif ( Diff >= 7 ) then
			Type = prefix .. FlagRSP_Locale_Estrong;
		elseif ( Diff >= 4 ) then
			Type = prefix .. FlagRSP_Locale_Vstrong;
		elseif ( Diff >= 2 ) then
			Type = prefix .. FlagRSP_Locale_Strong;
		elseif ( Diff >= -1 ) then
			Type = prefix .. FlagRSP_Locale_Equal;
		elseif ( Diff >= -GetQuestGreenRange()) then
			Type = prefix .. FlagRSP_Locale_Weak;
		elseif ( Diff >= (-4 -GetQuestGreenRange())) then
			Type = prefix .. FlagRSP_Locale_Puny;
		else
			Type = prefix .. FlagRSP_Locale_Epuny;
		end

		Type = Type .. "|r";
	else
		Type = "";
	end

	return Type;
end

--[[

- TooltipHandler:GetAlternateLevelText(playerID)

- Colors the level text the same as in the target frame.

]]--
function TooltipHandler:GetAlternateLevelText( playerID )
	local Type, prefix;
	local level;

	if string.sub(playerID,1,5) == "index" then
		local iname = self:GetCharacterName(playerID, "unit");
		level = TooltipHandler:GetFriendInfo("level", iname);
	else
		level = UnitLevel(playerID);
	end

	if level ~= nil and level ~= "" and level ~= 0 then

		local r,g,b;
		local color;

		color = GetQuestDifficultyColor(level);
		r = color.r;
		g = color.g;
		b = color.b;

		if ( level == -1 ) then
			level = "??"
			r=1;
			g=0.13;
			b=0.13;
		end

		prefix = "|cFF" .. FlagRSPStatic:GetHex(r, g, b);
		Type = prefix .. FlagRSP_Locale_TradLevelLine .. level;

		Type = Type .. "|r";
	else
		Type = "";
	end

	return Type;
end

function TooltipHandler:DecorateAddon( flagaddon )
    return flagaddon;
end


-- State recognition functions (friend, ignore, ...) --

--[[

- TooltipHandler:UpdateFriendEntries()

-- Updates the list of standard friend list entries.

]]--
function TooltipHandler:UpdateFriendEntries()
	self.friendEntries = nil;
	self.friendEntries = {};

	FlagRSP:PrintDebug("updating friend list.");
	for t = 1, GetNumFriends() do
		local fName = GetFriendInfo(t);
		self.friendEntries[t] = fName;
	end
end

--[[

- TooltipHandler:UpdateIgnoredList()

-- Updates the list of ignored players.

]]--
function TooltipHandler:UpdateIgnoredList()
	self.ignoredList = nil;
	self.ignoredList = {};

	FlagRSP:PrintDebug("updating ignored list.");
	for t = 1, GetNumIgnores() do
		local iName = GetIgnoreName(t);

		self.ignoredList[t] = iName;
	end
end

--[[

- TooltipHandler:UpdateGuildMembers()

-- Updates the list of own guild members.

]]--
function TooltipHandler:UpdateGuildMembers()
	self.guildMembers = nil;
	self.guildMembers = {};

	FlagRSP:PrintDebug("updating guild list.");
	local u = GetNumGuildMembers();
	for t = 1, u do
		local gName = GetGuildRosterInfo(t);

		self.guildMembers[t] = gName;
	end
end

--[[

- TooltipHandler:PlayerIsGuildMember(name)

- Checks, if player <name> is a guild member of the user.
- Returns true if so, false if not.

]]--
function TooltipHandler:PlayerIsGuildMember( name )
	local isMember = false;

	for i=1, table.getn(self.guildMembers) do
		if self.guildMembers[i] == name then
			isMember = true;
		end
	end

	return isMember;
end

--[[

- TooltipHandler:PlayerIsFriendEntry(name)

- Checks, if player <name> is a guild member of the user.
- Returns true if so, false if not.

]]--
function TooltipHandler:PlayerIsFriendEntry( name )
	local isFriend = false;

	for i=1, table.getn(self.friendEntries) do
		if self.friendEntries[i] == name then
			isFriend = true;
		end
	end

	return isFriend;
end

--[[

- TooltipHandler:PlayerIsIgnored(name)

- Checks, if player <name> is a guild member of the user.
- Returns true if so, false if not.

]]--
function TooltipHandler:PlayerIsIgnored( name )
	local isIgnored = false;

	for i=1, table.getn(self.ignoredList) do
		if self.ignoredList[i] == name then
			isIgnored = true;
		end
	end

	return isIgnored;
end

--[[

- TooltipHandler:PlayerIsKnown(name)

- Checks, if player name is known to us (self, friends, ignore, guild, Friendlist).
- Returns true if known, false if unknown.

]]--
function TooltipHandler:PlayerIsKnown( name )
	local isKnown = false;

	if name == UnitName("player") then
		isKnown = true;
	end

	if self:PlayerIsFriendEntry(name) then
		isKnown = true;
	end

	if 1 == 0 then
		for t = 1, GetNumFriends() do
			local fName = GetFriendInfo(t);
			if fName == name then
				isKnown = true;
				break;
			end
		end
	end

	if self:PlayerIsIgnored(name) then
		isKnown = true;
	end

	if 1 == 0 then
		for t = 1, GetNumIgnores() do
			local iName = GetIgnoreName(t);
			if iName == name then
				isKnown = true; 	
				break;
			end
		end
	end

	if self:PlayerIsGuildMember(name) then
		isKnown = true;
	end

	if 1 == 0 then
		local u = GetNumGuildMembers();

		for t = 1, u do
			local gName = GetGuildRosterInfo(t);
			if gName == name then
				isKnown = true;
				break;
			end
		end
	end

	if Friendlist:IsFriend(name) then
		isKnown = true;
	end

	return isKnown;
end


--[[

- TooltipHandler:PlayerIsIgnored(name)

- Checks, if player name is ignored.
- Returns true if ignored, false if not.

]]--
function TooltipHandler:IsPlayerIgnored(name)
	local isIgnored = false;

	for t = 1, GetNumIgnores() do
		local iName = GetIgnoreName(t);
		if iName == name then
			isIgnored = true; 	
			break;
		end
	end

	return isIgnored;
end



-- Tooltip Manipulation Help ---------------------

--[[

- TooltipHandler:GetTooltipBGColor(playerID)

- Returns background color for tooltip.

]]--
function TooltipHandler:GetTooltipBGColor( playerID )
	local r, g, b;
	r = 0.09;
	g = 0.09;
	b = 0.19;

	local reaction = UnitReaction(playerID, "player");

	if reaction ~= nil then
		if not UnitIsPlayer(playerID) and not UnitPlayerControlled(playerID) then
			if reaction >= 5 then r = 0.09; g = 0.19; b = 0.09; end
			if reaction < 4 then r = 0.10; g = 0.09; b = 0.09; end
			if reaction == 4 then r = 0.19; g = 0.19; b = 0.09; end
		elseif UnitIsPVP(playerID) then
			if reaction >= 5 then r = 0.09; g = 0.19; b = 0.1; end
			if reaction < 4 and not UnitIsPVP("player") then
				r = 0.19; g = 0.19; b = 0.09;
			elseif reaction < 4 and UnitIsPVP("player") then
				r = 0.19; g = 0.09; b = 0.09;
			end
			if reaction == 4 then r = 0.19; g = 0.19; b = 0.09; end
		else
			r = 0.09; g = 0.09; b = 0.19;
		end
	else
		if not UnitIsPlayer(playerID) and not UnitPlayerControlled(playerID) then
			r = 0.19; g = 0.19; b = 0.09;
		else
			r = 0.09; g = 0.09; b = 0.19;
		end
	end

	return r, g, b;
end

--[[

- TooltipHandler:GetTooltipNameColor(playerID)

- Returns text color for tooltip name line.

]]--
function TooltipHandler:GetTooltipNameColor( playerID )
	local r,g,b;

	r = 1;
	g = 1;
	b = 1;

	local reaction = UnitReaction(playerID, "player");

    -- Avatare:
    --    Normal: white, green (pvp)
    --    RP: violett?
    -- NPC:
    --    General: 2 (hostile, red), 4 (neutral, yellow), 5 (friendly, green)
    --    With Reputation: Reputation states and colors
	if reaction ~= nil then
		if not UnitIsPlayer(playerID) and not UnitPlayerControlled(playerID) then
			if reaction >= 5 then r = 0; g = 0.95; b = 0; end
			if reaction < 4 then r = 0.95; g = 0; b = 0; end
			if reaction == 4 then r = 0.95; g = 0.95; b = 0; end
		elseif UnitIsPVP(playerID) then
			if reaction >= 5 then r = 0; g = 0.95; b = 0; end
			if reaction < 4 and not UnitIsPVP("player") then
				r = 0.95; g = 0.95; b = 0.1;
			elseif reaction < 4 and UnitIsPVP("player") then
				r = 0.95; g = 0; b = 0;
			end
			if reaction == 4 then r = 0.95; g = 0.95; b = 0; end
		else
			r = 0.5; g = 0.5; b = 0.95;
		end
	else
		if not UnitIsPlayer(playerID) and not UnitPlayerControlled(playerID) then
			r = 0.95; g = 0.95; b = 0;
		else
			r = 0.5; g = 0.5; b = 0.95;
		end
	end

	return r,g,b;
end


-- Encapsulating/Interfaces/Foreign functions ----

function TooltipHandler:GetFlagInfo( tag, ingameName, realm )
    local flag;
    flag = FlagHandler:GetFlag(tag, ingameName, realm);
    return flag or "";
end

--[[ Takes care of getting the Friendlist information on <ingameName<-Realm>>
  Supported tags:
  type, index, class, name, rank, race,title, guild, notes, level, friendstate, surname
--]]
function TooltipHandler:GetFriendInfo( tag, ingameName, realm )
    tag = strlower(tag);

    local txt;
    
    if realm ~= nil and realm ~= FlagRSP.rName then
        ingameName = ingameName .. "-" .. realm;
    end

    if ( tag == "friendstate" ) then
        txt = Friendlist:GetFriendstateText(ingameName, realm);
    -- elseif ( tag == "type" ) then
    elseif ( tag == "class" ) then
        txt = Friendlist:GetClass(ingameName);
    elseif ( tag == "rank" ) then
        txt = Friendlist:GetRank(ingameName);
    elseif ( tag == "race" ) then
        txt = Friendlist:GetRace(ingameName);
    elseif ( tag == "title" ) then
        txt = Friendlist:GetTitle(ingameName);
    elseif ( tag == "guild" ) then
        txt = Friendlist:GetGuild(ingameName);
    elseif ( tag == "level" ) then
        txt = Friendlist:GetLevel(ingameName);
    elseif ( tag == "fullname") then
		-- local altN    = FriendlistHandler:GetInfo( "altName", ingameName );
		local firstN  = FriendlistHandler:GetInfo( "name"   , ingameName );
		local surname = FriendlistHandler:GetInfo( "surname", ingameName );
		
		-- if ( altN ~= "" or firstN ~= "" or surname ~= "" ) then
		if ( firstN ~= "" or surname ~= "" ) then
			-- txt = altN .. " " .. firstN .. " " .. surname;
			txt = firstN .. " " .. surname;
	    else
            -- txt = ingameName;
            txt = "";
		end
		txt = strtrim(txt);
    elseif ( tag == "surname" ) then
        txt = Friendlist:GetSurname(ingameName);
    elseif ( tag == "name" ) then
        txt = Friendlist:GetName(ingameName);
    elseif ( tag == "nameT" ) then
        txt = Friendlist:GetName(ingameName, true);
    elseif ( tag == "friendcolor" ) then
        txt = Friendlist:GetFriendstateColour(ingameName);
    end
    -- tag = FriendHandler:GetFlag(tag, ingameName, realm);
    return txt or "";
end